/*
 * Copyright 2008 Sony Corporation of America
 *
 *
 * Permission is hereby granted, free of charge, to any person obtaining
 * a copy of this Library and associated documentation files (the
 * "Library"), to deal in the Library without restriction, including
 * without limitation the rights to use, copy, modify, merge, publish,
 * distribute, sublicense, and/or sell copies of the Library, and to
 * permit persons to whom the Library is furnished to do so, subject to
 * the following conditions:
 *
 *  The above copyright notice and this permission notice shall be
 *  included in all copies or substantial portions of the Library.
 *
 *  If you modify the Library, you may copy and distribute your modified
 *  version of the Library in object code or as an executable provided
 *  that you also do one of the following:
 *
 *   Accompany the modified version of the Library with the complete
 *   corresponding machine-readable source code for the modified version
 *   of the Library; or,
 *
 *   Accompany the modified version of the Library with a written offer
 *   for a complete machine-readable copy of the corresponding source
 *   code of the modified version of the Library.
 *
 *
 * THE LIBRARY IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND,
 * EXPRESS OR IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF
 * MERCHANTABILITY, FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT.
 * IN NO EVENT SHALL THE AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY
 * CLAIM, DAMAGES OR OTHER LIABILITY, WHETHER IN AN ACTION OF CONTRACT,
 * TORT OR OTHERWISE, ARISING FROM, OUT OF OR IN CONNECTION WITH THE
 * LIBRARY OR THE USE OR OTHER DEALINGS IN THE LIBRARY.
 */

#ifndef MARS_TASK_QUEUE_H
#define MARS_TASK_QUEUE_H

/**
 * \file
 * \ingroup group_mars_task_queue
 * \brief <b>[host]</b> MARS Task Queue Flag API
 */

#include <stdint.h>
#include <mars/task_queue_types.h>

struct mars_context;

#if defined(__cplusplus)
extern "C" {
#endif

/**
 * \ingroup group_mars_task_queue
 * \brief <b>[host]</b> Creates a task queue.
 *
 * This function will allocate an instance of the task queue.
 * The queue allows for tasks to wait until data is available from a FIFO
 * data queue.
 * The queue should be used in pairs with various calls to push and pop the
 * the queue
 *
 * <b>Key Parameters</b>:
 * \n \n
 * \e size
 * - Specify the size of each data item in the queue.
 * - The size needs to be a multiple of 16 bytes and no larger than
 * \ref MARS_TASK_QUEUE_ENTRY_SIZE_MAX.
 *
 * \e depth
 * - Specify the depth of the total queue.
 * - The number specified here is the total number of data items the queue
 * can hold at a time.
 * - The size of the total queue is \e size * \e depth.
 *
 * \e direction
 * - Specify the communication direction of the queue.
 * - Must be one of \ref MARS_TASK_QUEUE_HOST_TO_MPU,
 * \ref MARS_TASK_QUEUE_MPU_TO_HOST, \ref MARS_TASK_QUEUE_MPU_TO_MPU.
 *
 * \param[in] mars		- pointer to MARS context
 * \param[out] queue		- address of pointer to queue instance
 * \param[in] size		- size of each data item in data buffer
 * \param[in] depth		- maximum number of data entries in data buffer
 * \param[in] direction		- direction of the event flag
 * \return
 *	MARS_SUCCESS		- successfully created queue
 * \n	MARS_ERROR_NULL		- null pointer specified
 * \n	MARS_ERROR_PARAMS	- size is not multiple of 16 bytes or is greater
 *				  than 16KB maximum
 * \n	MARS_ERROR_PARAMS	- depth exceeds allowed limit
 * \n	MARS_ERROR_PARAMS	- invalid direction specified
 */
int mars_task_queue_create(struct mars_context *mars,
				struct mars_task_queue **queue,
				uint32_t size,
				uint32_t depth,
				uint8_t direction);

/**
 * \ingroup group_mars_task_queue
 * \brief <b>[host]</b> Destroys a task queue.
 *
 * This function will free any resources allocated during creation of the task
 * queue.
 *
 * \param[in] queue		- pointer to queue instance
 * \return
 *	MARS_SUCCESS		- successfully destroyed queue
 * \n	MARS_ERROR_NULL		- null pointer specified
 * \n	MARS_ERROR_ALIGN	- instance not properly aligned
 */
int mars_task_queue_destroy(struct mars_task_queue *queue);

/**
 * \ingroup group_mars_task_queue
 * \brief <b>[host]</b> Returns the number of data items in the task queue.
 *
 * This function will return the total number of data items in the queue at the
 * time of the call.
 *
 * \param[in] queue		- pointer to queue instance
 * \param[out] count		- pointer to variable to store return count
 * \return
 *	MARS_SUCCESS		- successfully returned count
 * \n	MARS_ERROR_NULL		- null pointer specified
 * \n	MARS_ERROR_ALIGN	- instance not aligned properly
 */
int mars_task_queue_count(struct mars_task_queue *queue, uint32_t *count);

/**
 * \ingroup group_mars_task_queue
 * \brief <b>[host]</b> Clears the data items in the task queue.
 *
 * This function will clear all data items currently in the queue.
 * Tasks waiting to pop data from the queue will remain waiting.
 * Tasks waiting to push data into the queue will resume.
 *
 * \param[in] queue		- pointer to queue instance
 * \return
 *	MARS_SUCCESS		- successfully cleared queue
 * \n	MARS_ERROR_NULL		- null pointer specified
 * \n	MARS_ERROR_ALIGN	- instance not aligned properly
 */
int mars_task_queue_clear(struct mars_task_queue *queue);

/**
 * \ingroup group_mars_task_queue
 * \brief <b>[host]</b> Pushes the data specified into the task queue.
 *
 * This function will push the data specified into the queue.
 * The task waiting longest to pop data from the queue can resume.
 *
 * If the queue is full at the time of this call, this function will block until
 * there is room in the queue to push the data.
 *
 * <b>Key Parameters</b>:
 * \n \n
 * \e data
 * - Specify the pointer to user data to copy into the queue.
 * - The size of data should be equal to the size specified at queue creation.
 *
 * \param[in] queue		- pointer to queue instance
 * \param[in] data		- address of data to be pushed into queue
 * \return
 *	MARS_SUCCESS		- successfully pushed data into queue
 * \n	MARS_ERROR_NULL		- null pointer specified
 * \n	MARS_ERROR_ALIGN	- instance not aligned properly
 * \n	MARS_ERROR_STATE	- invalid direction
 */
int mars_task_queue_push(struct mars_task_queue *queue, const void *data);

/**
 * \ingroup group_mars_task_queue
 * \brief <b>[host]</b> Pushes the data specified into the task queue.
 *
 * This function will push the data specified into the queue.
 * The task waiting longest to pop data from the queue can resume.
 *
 * If the queue is full at the time of this call, this function will not block
 * and returns immediately with \ref MARS_ERROR_BUSY.
 *
 * <b>Key Parameters</b>:
 * \n \n
 * \e data
 * - Specify the pointer to user data to copy into the queue.
 * - The size of data should be equal to the size specified at queue creation.
 *
 * \param[in] queue		- pointer to queue instance
 * \param[in] data		- address of data to be pushed into queue
 * \return
 *	MARS_SUCCESS		- successfully pushed data into queue
 * \n	MARS_ERROR_NULL		- null pointer specified
 * \n	MARS_ERROR_ALIGN	- instance not aligned properly
 * \n	MARS_ERROR_BUSY		- queue is full
 * \n	MARS_ERROR_STATE	- invalid direction
 */
int mars_task_queue_try_push(struct mars_task_queue *queue, const void *data);

/**
 * \ingroup group_mars_task_queue
 * \brief <b>[host]</b> Pops data from a task queue.
 *
 * This function will pop data from the queue.
 *
 * If the queue is empty at the time of this call, this function will block
 * until there is data to be popped from the queue.
 *
 * <b>Key Parameters</b>:
 * \n \n
 * \e data
 * - Specify the pointer to some allocated memory to copy the popped data.
 * - The size of the allocated memory area should be equal to the size specified
 * at queue creation.
 *
 * \param[in] queue		- pointer to task instance
 * \param[in] data		- address of data to be popped from queue
 * \return
 *	MARS_SUCCESS		- successfully received event
 * \n	MARS_ERROR_NULL		- null pointer specified
 * \n	MARS_ERROR_ALIGN	- instance not aligned properly
 * \n	MARS_ERROR_STATE	- invalid direction
 */
int mars_task_queue_pop(struct mars_task_queue *queue, void *data);

/**
 * \ingroup group_mars_task_queue
 * \brief <b>[host]</b> Pops data from a task queue.
 *
 * This function will pop data from the queue.
 *
 * If the queue is empty at the time of this call, this function will not block
 * and returns immediately with \ref MARS_ERROR_BUSY.
 *
 * <b>Key Parameters</b>:
 * \n \n
 * \e data
 * - Specify the pointer to some allocated memory to copy the popped data.
 * - The size of the allocated memory area should be equal to the size specified
 * at queue creation.
 *
 * \param[in] queue		- pointer to task instance
 * \param[in] data		- address of data to be popped from queue
 * \return
 *	MARS_SUCCESS		- successfully received event
 * \n	MARS_ERROR_NULL		- null pointer specified
 * \n	MARS_ERROR_ALIGN	- instance not aligned properly
 * \n	MARS_ERROR_BUSY		- queue is empty
 * \n	MARS_ERROR_STATE	- invalid direction
 */
int mars_task_queue_try_pop(struct mars_task_queue *queue, void *data);

/**
 * \ingroup group_mars_task_queue
 * \brief <b>[host]</b> Pops data from a task queue without removing it.
 *
 * This function will retrieve data from the queue without removing it from the
 * queue.
 *
 * If the queue is empty at the time of this call, this function will block
 * until there is data to be retrieved from the queue.
 *
 * <b>Key Parameters</b>:
 * \n \n
 * \e data
 * - Specify the pointer to some allocated memory to copy the popped data.
 * - The size of the allocated memory area should be equal to the size specified
 * at queue creation.
 *
 * \param[in] queue		- pointer to task instance
 * \param[in] data		- address of data to be popped from queue
 * \return
 *	MARS_SUCCESS		- successfully received event
 * \n	MARS_ERROR_NULL		- null pointer specified
 * \n	MARS_ERROR_ALIGN	- instance not aligned properly
 * \n	MARS_ERROR_STATE	- invalid direction
 */
int mars_task_queue_peek(struct mars_task_queue *queue, void *data);

/**
 * \ingroup group_mars_task_queue
 * \brief <b>[host]</b> Pops data from a task queue without removing it.
 *
 * This function will retrieve data from the queue.
 *
 * If the queue is empty at the time of this call, this function will not block
 * and returns immediately with \ref MARS_ERROR_BUSY.
 *
 * <b>Key Parameters</b>:
 * \n \n
 * \e data
 * - Specify the pointer to some allocated memory to copy the popped data.
 * - The size of the allocated memory area should be equal to the size specified
 * at queue creation.
 *
 * \param[in] queue		- pointer to task instance
 * \param[in] data		- address of data to be popped from queue
 * \return
 *	MARS_SUCCESS		- successfully received event
 * \n	MARS_ERROR_NULL		- null pointer specified
 * \n	MARS_ERROR_ALIGN	- instance not aligned properly
 * \n	MARS_ERROR_BUSY		- queue is empty
 * \n	MARS_ERROR_STATE	- invalid direction
 */
int mars_task_queue_try_peek(struct mars_task_queue *queue, void *data);

#if defined(__cplusplus)
}
#endif

#endif
