/*
 * Copyright 2008 Sony Corporation of America
 *
 *
 * Permission is hereby granted, free of charge, to any person obtaining
 * a copy of this Library and associated documentation files (the
 * "Library"), to deal in the Library without restriction, including
 * without limitation the rights to use, copy, modify, merge, publish,
 * distribute, sublicense, and/or sell copies of the Library, and to
 * permit persons to whom the Library is furnished to do so, subject to
 * the following conditions:
 *
 *  The above copyright notice and this permission notice shall be
 *  included in all copies or substantial portions of the Library.
 *
 *  If you modify the Library, you may copy and distribute your modified
 *  version of the Library in object code or as an executable provided
 *  that you also do one of the following:
 *
 *   Accompany the modified version of the Library with the complete
 *   corresponding machine-readable source code for the modified version
 *   of the Library; or,
 *
 *   Accompany the modified version of the Library with a written offer
 *   for a complete machine-readable copy of the corresponding source
 *   code of the modified version of the Library.
 *
 *
 * THE LIBRARY IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND,
 * EXPRESS OR IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF
 * MERCHANTABILITY, FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT.
 * IN NO EVENT SHALL THE AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY
 * CLAIM, DAMAGES OR OTHER LIABILITY, WHETHER IN AN ACTION OF CONTRACT,
 * TORT OR OTHERWISE, ARISING FROM, OUT OF OR IN CONNECTION WITH THE
 * LIBRARY OR THE USE OR OTHER DEALINGS IN THE LIBRARY.
 */

#include <malloc.h>
#include <string.h>
#include <pthread.h>
#include <libspe2.h>

#include "mars/context.h"
#include "mars/core.h"
#include "mars/error.h"

#include "context_internal_types.h"
#include "kernel_internal_types.h"
#include "workload_queue.h"

extern struct spe_program_handle mars_kernel_entry;
static struct mars_context *mars_shared_context;
static pthread_mutex_t mars_mutex = PTHREAD_MUTEX_INITIALIZER;

static void *mpu_context_thread(void *arg)
{
	int ret;
	unsigned int entry = SPE_DEFAULT_ENTRY;
	struct mars_kernel_params *params = (struct mars_kernel_params *)arg;
	struct spe_context *spe;

	spe = spe_context_create(0, NULL);
	if (!spe)
		return (void *)MARS_ERROR_INTERNAL;

	ret = spe_program_load(spe, &mars_kernel_entry);
	if (ret) {
		spe_context_destroy(spe);
		return (void *)MARS_ERROR_INTERNAL;
	}

	ret = spe_context_run(spe, &entry, 0, params, NULL, NULL);
	if (ret) {
		spe_context_destroy(spe);
		return (void *)MARS_ERROR_INTERNAL;
	}

	ret = spe_context_destroy(spe);
	if (ret)
		return (void *)MARS_ERROR_INTERNAL;

	return (void *)MARS_SUCCESS;
}

static int mpu_contexts_create(struct mars_context *mars, uint32_t num_mpus)
{
	int ret;
	unsigned int i;

	/* create threads for each mpu context */
	for (i = mars->mpu_context_count; i < num_mpus; i++) {
		struct mars_kernel_params *params = &mars->kernel_params[i];

		params->id = i;
		params->mars_context_ea =
			mars_ptr_to_ea(mars);
		params->workload_queue_ea =
			mars_ptr_to_ea(mars->workload_queue);

		ret = pthread_create(&mars->mpu_context_threads[i], NULL,
			mpu_context_thread, params);
		if (ret)
			return MARS_ERROR_INTERNAL;

		mars->mpu_context_count++;
	}

	return MARS_SUCCESS;
}

static int mpu_contexts_destroy(struct mars_context *mars)
{
	int ret;
	void *p_ret;
	unsigned int i;

	/* shutdown the workload queue so mpu context threads exit */
	ret = mars_workload_queue_exit(mars->workload_queue);
	if (ret != MARS_SUCCESS)
		return ret;

	/* join all mpu context threads */
	for (i = 0; i < mars->mpu_context_count; i++) {
		ret = pthread_join(mars->mpu_context_threads[i], &p_ret);
		if (ret || p_ret)
			return MARS_ERROR_INTERNAL;
	}

	return MARS_SUCCESS;
}

int mars_context_create(struct mars_context **mars_ret, uint32_t num_mpus,
			uint8_t shared)
{
	int ret;
	int num_mpus_max;
	struct mars_context *mars = NULL;

	/* check function params */
	if (!mars_ret)
		return MARS_ERROR_NULL;

	/* check number of mpus to use */
	num_mpus_max = spe_cpu_info_get(SPE_COUNT_PHYSICAL_SPES, -1);
	if (num_mpus_max <= 0)
		return MARS_ERROR_INTERNAL;
	if (num_mpus_max < (int)num_mpus)
		return MARS_ERROR_PARAMS;
	if (!num_mpus)
		num_mpus = num_mpus_max;

	/* lock mutex */
	ret = pthread_mutex_lock(&mars_mutex);
	if (ret)
		return MARS_ERROR_INTERNAL;

	/* shared context requested */
	if (shared && mars_shared_context) {
		/* create any extra mpu contexts necessary */
		ret = mpu_contexts_create(mars_shared_context, num_mpus);
		if (ret != MARS_SUCCESS)
			goto error;

		/* increment shared context reference count */
		mars_shared_context->reference_count++;

		/* return the shared context */
		*mars_ret = mars_shared_context;

		goto done;
	}

	/* allocate */
	mars = malloc(sizeof(struct mars_context));
	if (!mars) {
		ret = MARS_ERROR_MEMORY;
		goto error;
	}

	/* zero context */
	memset(mars, 0, sizeof(struct mars_context));

	/* increment reference count */
	mars->reference_count++;

	/* allocate kernel params */
	mars->kernel_params = (struct mars_kernel_params *)
		memalign(MARS_KERNEL_PARAMS_ALIGN,
			sizeof(struct mars_kernel_params) * num_mpus_max);
	if (!mars->kernel_params) {
		ret = MARS_ERROR_MEMORY;
		goto error_malloc_kernel_params;
	}

	/* allocate mpu context thread array */
	mars->mpu_context_threads = (pthread_t *)
		malloc(sizeof(pthread_t) * num_mpus_max);
	if (!mars->mpu_context_threads) {
		ret = MARS_ERROR_MEMORY;
		goto error_malloc_mpu_context_threads;
	}

	/* create workload queue */
	ret = mars_workload_queue_create(&mars->workload_queue);
	if (ret != MARS_SUCCESS)
		goto error_workload_queue_create;

	/* create mpu contexts */
	ret = mpu_contexts_create(mars, num_mpus);
	if (ret != MARS_SUCCESS)
		goto error_mpu_contexts_create;

	/* set the shared context pointer */
	if (shared)
		mars_shared_context = mars;

	/* return mars context pointer */
	*mars_ret = mars;

done:
	/* unlock mutex */
	ret = pthread_mutex_unlock(&mars_mutex);
	if (ret) {
		ret = MARS_ERROR_INTERNAL;
		goto error_pthread_mutex_unlock;
	}

	return MARS_SUCCESS;

error_pthread_mutex_unlock:
	mpu_contexts_destroy(mars);
error_mpu_contexts_create:
	mars_workload_queue_destroy(mars->workload_queue);
error_workload_queue_create:
	free(mars->mpu_context_threads);
error_malloc_mpu_context_threads:
	free(mars->kernel_params);
error_malloc_kernel_params:
	free(mars);
error:
	pthread_mutex_unlock(&mars_mutex);

	return ret;
}

int mars_context_destroy(struct mars_context *mars)
{
	int ret;

	/* check function params */
	if (!mars)
		return MARS_ERROR_NULL;

	/* lock mutex */
	ret = pthread_mutex_lock(&mars_mutex);
	if (ret)
		return MARS_ERROR_INTERNAL;

	/* decrement reference count */
	mars->reference_count--;

	/* reference count is not 0 so return */
	if (mars->reference_count)
		goto done;

	/* destroy mpu contexts */
	if (mars->mpu_context_count) {
		ret = mpu_contexts_destroy(mars);
		if (ret != MARS_SUCCESS)
			goto error;
	}

	/* destroy workload queue */
	if (mars->workload_queue) {
		ret = mars_workload_queue_destroy(mars->workload_queue);
		if (ret != MARS_SUCCESS)
			goto error;
	}

	/* free allocated memory */
	free(mars->mpu_context_threads);
	free(mars->kernel_params);
	free(mars);

	/* check if it is the shared context pointer and set to NULL */
	if (mars == mars_shared_context)
		mars_shared_context = NULL;

done:
	/* unlock mutex */
	ret = pthread_mutex_unlock(&mars_mutex);
	if (ret)
		return MARS_ERROR_INTERNAL;

	return MARS_SUCCESS;

error:
	pthread_mutex_unlock(&mars_mutex);

	return ret;
}
