/*
 * Copyright 2008 Sony Corporation of America
 *
 *
 * Permission is hereby granted, free of charge, to any person obtaining
 * a copy of this Library and associated documentation files (the
 * "Library"), to deal in the Library without restriction, including
 * without limitation the rights to use, copy, modify, merge, publish,
 * distribute, sublicense, and/or sell copies of the Library, and to
 * permit persons to whom the Library is furnished to do so, subject to
 * the following conditions:
 *
 *  The above copyright notice and this permission notice shall be
 *  included in all copies or substantial portions of the Library.
 *
 *  If you modify the Library, you may copy and distribute your modified
 *  version of the Library in object code or as an executable provided
 *  that you also do one of the following:
 *
 *   Accompany the modified version of the Library with the complete
 *   corresponding machine-readable source code for the modified version
 *   of the Library; or,
 *
 *   Accompany the modified version of the Library with a written offer
 *   for a complete machine-readable copy of the corresponding source
 *   code of the modified version of the Library.
 *
 *
 * THE LIBRARY IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND,
 * EXPRESS OR IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF
 * MERCHANTABILITY, FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT.
 * IN NO EVENT SHALL THE AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY
 * CLAIM, DAMAGES OR OTHER LIABILITY, WHETHER IN AN ACTION OF CONTRACT,
 * TORT OR OTHERWISE, ARISING FROM, OUT OF OR IN CONNECTION WITH THE
 * LIBRARY OR THE USE OR OTHER DEALINGS IN THE LIBRARY.
 */

#include <elf.h>
#include <string.h>
#include <malloc.h>

#include "mars/task.h"
#include "mars/core.h"
#include "mars/error.h"

#include "task_internal_types.h"
#include "context_internal_types.h"
#include "workload_queue.h"

const struct mars_task_context_save_unit mars_task_context_save_all[] =
{
	{ .addr = 0, .size = MARS_TASK_CONTEXT_SAVE_SIZE_MAX },
	{ .addr = 0, .size = 0 },
};

int mars_task_create(struct mars_context *mars,
		struct mars_task_id *id_ret,
		const char *name, const void *elf_image,
		const struct mars_task_context_save_unit *context_save_unit)
{
	int ret;
	Elf32_Ehdr *ehdr;
	Elf32_Phdr *phdr;
	uint16_t workload_id;
	struct mars_task_context *task;
	struct mars_workload_context *workload;

	int context_save_area_size = 0;
	int context_save_unit_count = 1;
	const struct mars_task_context_save_unit *p = context_save_unit;

	/* check function params */
	if (!mars)
		return MARS_ERROR_NULL;
	if (!id_ret)
		return MARS_ERROR_NULL;
	if (!elf_image)
		return MARS_ERROR_NULL;
	if (name && strlen(name) > MARS_TASK_NAME_LEN_MAX)
		return MARS_ERROR_PARAMS;

	/* process elf header information */
	ehdr = (Elf32_Ehdr *)elf_image;
	phdr = (Elf32_Phdr *)((void *)ehdr + ehdr->e_phoff);

	if (phdr->p_type != PT_LOAD)
		return MARS_ERROR_FORMAT;
	if (phdr->p_vaddr != MARS_TASK_BASE_ADDR)
		return MARS_ERROR_FORMAT;

	/* begin process to add the task to the workload queue */
	ret = mars_workload_queue_add_begin(mars->workload_queue, &workload_id,
		MARS_WORKLOAD_TYPE_TASK, &workload);
	if (ret != MARS_SUCCESS)
		return ret;

	/* cast workload context to task context */
	task = (struct mars_task_context *)workload;

	/* initialize task id */
	task->id.mars_context_ea = mars_ptr_to_ea(mars);
	task->id.workload_id = workload_id;
	if (name)
		strcpy((char *)task->id.name, name);
	else
		task->id.name[0] = 0;

	/* initialize the elf parameters */
	task->exec_ea = mars_ptr_to_ea((void *)ehdr + phdr->p_offset);
	task->exec_size = phdr->p_filesz;
	task->bss_size = phdr->p_memsz - phdr->p_filesz;
	task->vaddr = phdr->p_vaddr;
	task->entry = ehdr->e_entry;

	/* no context save - run complete */
	if (!context_save_unit) {
		task->context_save_unit_ea = 0;
		task->context_save_area_ea = 0;
		goto done;
	}

	/* calculate save unit count and save area size */
	while (p->size) {
		if (p->addr & MARS_TASK_CONTEXT_SAVE_UNIT_ADDR_ALIGN_MASK ||
			p->size & MARS_TASK_CONTEXT_SAVE_UNIT_SIZE_ALIGN_MASK) {
			ret = MARS_ERROR_PARAMS;
			goto error_context_save_unit_addr_align;
		}

		context_save_area_size += context_save_unit->size;
		context_save_unit_count++;
		p++;

		if (context_save_unit_count == MARS_TASK_CONTEXT_SAVE_UNIT_MAX)
			break;
	}

	/* allocate context save unit storage */
	task->context_save_unit_ea = mars_ptr_to_ea(memalign(
		MARS_TASK_CONTEXT_SAVE_UNIT_ALIGN,
		context_save_unit_count * MARS_TASK_CONTEXT_SAVE_UNIT_SIZE));
	if (!task->context_save_unit_ea) {
		ret = MARS_ERROR_MEMORY;
		goto error_malloc_context_save_unit;
	}

	/* allocate context save area */
	task->context_save_area_ea = mars_ptr_to_ea(memalign(
		MARS_TASK_CONTEXT_SAVE_ALIGN, context_save_area_size));
	if (!task->context_save_area_ea) {
		ret = MARS_ERROR_MEMORY;
		goto error_malloc_context_save_area;
	}

	/* copy the context save struct into allocated storage */
	memcpy(mars_ea_to_ptr(task->context_save_unit_ea),
		context_save_unit,
		context_save_unit_count * MARS_TASK_CONTEXT_SAVE_UNIT_SIZE);

done:
	/* end process to add the task to the workload queue */
	ret = mars_workload_queue_add_end(mars->workload_queue, workload_id);
	if (ret != MARS_SUCCESS)
		goto error_workload_queue_add_end;

	/* return id to caller */
	*id_ret = task->id;

	return MARS_SUCCESS;

error_workload_queue_add_end:
	free(mars_ea_to_ptr(task->context_save_area_ea));
error_malloc_context_save_area:
	free(mars_ea_to_ptr(task->context_save_unit_ea));
error_malloc_context_save_unit:
error_context_save_unit_addr_align:
	mars_workload_queue_add_cancel(mars->workload_queue, workload_id);

	return ret;
}

int mars_task_destroy(struct mars_task_id *id)
{
	int ret;
	struct mars_context *mars;
	struct mars_task_context *task;
	struct mars_workload_context *workload;

	/* check function params */
	if (!id)
		return MARS_ERROR_NULL;
	if (!id->mars_context_ea)
		return MARS_ERROR_PARAMS;
	if (!id->workload_id >= MARS_WORKLOAD_MAX)
		return MARS_ERROR_PARAMS;

	/* get mars context pointer from task id */
	mars = mars_ea_to_ptr(id->mars_context_ea);

	/* begin process to remove the task from the workload queue */
	ret = mars_workload_queue_remove_begin(mars->workload_queue,
		id->workload_id, &workload);
	if (ret != MARS_SUCCESS)
		return ret;

	/* cast workload context to task context */
	task = (struct mars_task_context *)workload;

	/* free the allocated context save area if it has one */
	if (task->context_save_area_ea)
		free(mars_ea_to_ptr(task->context_save_area_ea));

	/* free the allocated context save unit storage if it has one */
	if (task->context_save_unit_ea)
		free(mars_ea_to_ptr(task->context_save_unit_ea));

	/* end process to remove the task from the workload queue */
	ret = mars_workload_queue_remove_end(mars->workload_queue,
		id->workload_id);
	if (ret != MARS_SUCCESS)
		return ret;

	return MARS_SUCCESS;
}

int mars_task_schedule(struct mars_task_id *id, struct mars_task_args *args,
			uint8_t priority)
{
	int ret;
	struct mars_context *mars;
	struct mars_task_context *task;
	struct mars_workload_context *workload;

	/* check function params */
	if (!id)
		return MARS_ERROR_NULL;
	if (!id->mars_context_ea)
		return MARS_ERROR_PARAMS;
	if (!id->workload_id >= MARS_WORKLOAD_MAX)
		return MARS_ERROR_PARAMS;

	/* get mars context pointer from task id */
	mars = mars_ea_to_ptr(id->mars_context_ea);

	/* begin process to schedule the workload in the workload queue */
	ret = mars_workload_queue_schedule_begin(mars->workload_queue,
		id->workload_id, priority, &workload);
	if (ret != MARS_SUCCESS)
		return ret;

	/* cast workload context to task context */
	task = (struct mars_task_context *)workload;

	/* initialize task specific context variables */
	task->stack = 0;
	if (args)
		memcpy(&task->args, args, sizeof(struct mars_task_args));

	/* end process to schedule the workload in the workload queue */
	ret = mars_workload_queue_schedule_end(mars->workload_queue,
		id->workload_id);
	if (ret != MARS_SUCCESS)
		return ret;

	return MARS_SUCCESS;
}

int mars_task_wait(struct mars_task_id *id, int32_t *exit_code)
{
	int ret;
	struct mars_context *mars;
	struct mars_task_context *task;
	struct mars_workload_context *workload;

	/* check function params */
	if (!id)
		return MARS_ERROR_NULL;
	if (!id->mars_context_ea)
		return MARS_ERROR_PARAMS;
	if (!id->workload_id >= MARS_WORKLOAD_MAX)
		return MARS_ERROR_PARAMS;

	/* get mars context pointer from task id */
	mars = mars_ea_to_ptr(id->mars_context_ea);

	/* blocking wait for workload completion */
	ret = mars_workload_queue_wait(mars->workload_queue, id->workload_id,
		&workload);
	if (ret != MARS_SUCCESS)
		return ret;

	/* cast workload context to task context */
	task = (struct mars_task_context *)workload;

	/* exit_code requested so return it to caller */
	if (exit_code)
		*exit_code = task->exit_code;

	return MARS_SUCCESS;
}

int mars_task_try_wait(struct mars_task_id *id, int32_t *exit_code)
{
	int ret;
	struct mars_context *mars;
	struct mars_task_context *task;
	struct mars_workload_context *workload;

	/* check function params */
	if (!id)
		return MARS_ERROR_NULL;
	if (!id->mars_context_ea)
		return MARS_ERROR_PARAMS;
	if (!id->workload_id >= MARS_WORKLOAD_MAX)
		return MARS_ERROR_PARAMS;

	/* get mars context pointer from task id */
	mars = mars_ea_to_ptr(id->mars_context_ea);

	/* non-blocking wait for workload completion */
	ret = mars_workload_queue_try_wait(mars->workload_queue,
		id->workload_id, &workload);
	if (ret != MARS_SUCCESS)
		return ret;

	/* cast workload context to task context */
	task = (struct mars_task_context *)workload;

	/* exit_code requested so return it to caller */
	if (exit_code)
		*exit_code = task->exit_code;

	return MARS_SUCCESS;
}
