/*
 * Copyright 2008 Sony Corporation of America
 *
 *
 * Permission is hereby granted, free of charge, to any person obtaining
 * a copy of this Library and associated documentation files (the
 * "Library"), to deal in the Library without restriction, including
 * without limitation the rights to use, copy, modify, merge, publish,
 * distribute, sublicense, and/or sell copies of the Library, and to
 * permit persons to whom the Library is furnished to do so, subject to
 * the following conditions:
 *
 *  The above copyright notice and this permission notice shall be
 *  included in all copies or substantial portions of the Library.
 *
 *  If you modify the Library, you may copy and distribute your modified
 *  version of the Library in object code or as an executable provided
 *  that you also do one of the following:
 *
 *   Accompany the modified version of the Library with the complete
 *   corresponding machine-readable source code for the modified version
 *   of the Library; or,
 *
 *   Accompany the modified version of the Library with a written offer
 *   for a complete machine-readable copy of the corresponding source
 *   code of the modified version of the Library.
 *
 *
 * THE LIBRARY IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND,
 * EXPRESS OR IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF
 * MERCHANTABILITY, FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT.
 * IN NO EVENT SHALL THE AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY
 * CLAIM, DAMAGES OR OTHER LIABILITY, WHETHER IN AN ACTION OF CONTRACT,
 * TORT OR OTHERWISE, ARISING FROM, OUT OF OR IN CONNECTION WITH THE
 * LIBRARY OR THE USE OR OTHER DEALINGS IN THE LIBRARY.
 */

#include <sched.h>
#include <string.h>
#include <malloc.h>

#include "mars/task_queue.h"
#include "mars/core.h"
#include "mars/mutex.h"
#include "mars/error.h"

#include "task_queue_internal_types.h"
#include "context_internal_types.h"
#include "mutex_internal_types.h"
#include "workload_queue.h"

int mars_task_queue_create(struct mars_context *mars,
				struct mars_task_queue **queue_ret,
				uint32_t size,
				uint32_t depth,
				uint8_t direction)
{
	struct mars_task_queue *queue;
	void *buffer;

	/* check function params */
	if (!mars)
		return MARS_ERROR_NULL;
	if (!queue_ret)
		return MARS_ERROR_NULL;
	if (size & MARS_TASK_QUEUE_ENTRY_SIZE_MASK)
		return MARS_ERROR_PARAMS;
	if (size > MARS_TASK_QUEUE_ENTRY_SIZE_MAX)
		return MARS_ERROR_PARAMS;
	if (direction != MARS_TASK_QUEUE_HOST_TO_MPU &&
		direction != MARS_TASK_QUEUE_MPU_TO_HOST &&
		direction != MARS_TASK_QUEUE_MPU_TO_MPU)
		return MARS_ERROR_PARAMS;

	/* allocate queue instance */
	queue = (struct mars_task_queue *)
		memalign(MARS_TASK_QUEUE_ALIGN, MARS_TASK_QUEUE_SIZE);
	if (!queue)
		return MARS_ERROR_MEMORY;

	/* allocate queue buffer instance */
	buffer = memalign(MARS_TASK_QUEUE_BUFFER_ALIGN, size * depth);
	if (!buffer)
		return MARS_ERROR_MEMORY;

	mars_mutex_reset((struct mars_mutex *)queue);

	/* initialize queue instance */
	queue->mars_context_ea = mars_ptr_to_ea(mars);
	queue->buffer_ea = mars_ptr_to_ea(buffer);
	queue->push_ea = mars_ptr_to_ea(buffer);
	queue->pop_ea = mars_ptr_to_ea(buffer);
	queue->size = size;
	queue->depth = depth;
	queue->direction = direction;
	queue->count = 0;
	queue->push_wait_count = 0;
	queue->pop_wait_count = 0;

	/* return queue instance pointer */
	*queue_ret = queue;

	return MARS_SUCCESS;
}

int mars_task_queue_destroy(struct mars_task_queue *queue)
{
	/* check function params */
	if (!queue)
		return MARS_ERROR_NULL;

	free(mars_ea_to_ptr(queue->buffer_ea));
	free(queue);

	return MARS_SUCCESS;
}

int mars_task_queue_count(struct mars_task_queue *queue, uint32_t *count)
{
	/* check function params */
	if (!queue)
		return MARS_ERROR_NULL;
	if (!count)
		return MARS_ERROR_NULL;
	if ((uintptr_t)queue & MARS_TASK_QUEUE_ALIGN_MASK)
		return MARS_ERROR_ALIGN;

	mars_mutex_lock((struct mars_mutex *)queue);

	/* return current items in queue */
	*count = queue->count;

	mars_mutex_unlock((struct mars_mutex *)queue);

	return MARS_SUCCESS;
}

int mars_task_queue_clear(struct mars_task_queue *queue)
{
	int i;
	struct mars_context *mars;

	/* check function params */
	if (!queue)
		return MARS_ERROR_NULL;
	if ((uintptr_t)queue & MARS_TASK_QUEUE_ALIGN_MASK)
		return MARS_ERROR_ALIGN;

	mars_mutex_lock((struct mars_mutex *)queue);

	/* get mars context pointer */
	mars = mars_ea_to_ptr(queue->mars_context_ea);

	/* signal all waiting tasks that queue is ready for push */
	for (i = 0; i < queue->push_wait_count; i++)
		mars_workload_queue_signal_send(mars->workload_queue,
			queue->push_wait_id[0]);

	/* flush all ids from push wait list */
	queue->push_wait_count = 0;

	/* clear the queue count and push/pop ea */
	queue->count = 0;
	queue->push_ea = queue->buffer_ea;
	queue->pop_ea = queue->buffer_ea;

	mars_mutex_unlock((struct mars_mutex *)queue);

	return MARS_SUCCESS;
}

static void push_update(struct mars_task_queue *queue)
{
	struct mars_context *mars = mars_ea_to_ptr(queue->mars_context_ea);

	/* signal waiting task that queue is ready for pop */
	if (queue->pop_wait_count) {
		/* signal waiting task */
		mars_workload_queue_signal_send(mars->workload_queue,
			queue->pop_wait_id[0]);

		/* flush id from pop wait list */
		queue->pop_wait_count--;
		memmove(&queue->pop_wait_id[0], &queue->pop_wait_id[1],
			sizeof(uint16_t) * queue->pop_wait_count);
	}

	/* increment queue count */
	queue->count++;

	/* increment queue push ea */
	queue->push_ea += queue->size;

	/* wrap to front of queue if necessary */
	if (queue->push_ea == queue->buffer_ea + (queue->size * queue->depth))
		queue->push_ea = queue->buffer_ea;
}

static int push(struct mars_task_queue *queue, const void *data, int try)
{
	/* check function params */
	if (!queue)
		return MARS_ERROR_NULL;
	if ((uintptr_t)queue & MARS_TASK_QUEUE_ALIGN_MASK)
		return MARS_ERROR_ALIGN;
	if (queue->direction != MARS_TASK_QUEUE_HOST_TO_MPU)
		return MARS_ERROR_STATE;

	mars_mutex_lock((struct mars_mutex *)queue);

	/* queue is full so wait */
	while (queue->count == queue->depth) {
		mars_mutex_unlock((struct mars_mutex *)queue);

		/* only try so return busy */
		if (try)
			return MARS_ERROR_BUSY;

		while (queue->count == queue->depth)
			sched_yield();

		mars_mutex_lock((struct mars_mutex *)queue);
	}

	/* copy data into queue */
	memcpy(mars_ea_to_ptr(queue->push_ea), data, queue->size);

	/* update queue data */
	push_update(queue);

	mars_mutex_unlock((struct mars_mutex *)queue);

	return MARS_SUCCESS;
}

int mars_task_queue_push(struct mars_task_queue *queue, const void *data)
{
	return push(queue, data, 0);
}

int mars_task_queue_try_push(struct mars_task_queue *queue, const void *data)
{
	return push(queue, data, 1);
}

static void pop_update(struct mars_task_queue *queue)
{
	struct mars_context *mars = mars_ea_to_ptr(queue->mars_context_ea);

	/* signal waiting task that queue is ready for push */
	if (queue->push_wait_count) {
		/* signal waiting task */
		mars_workload_queue_signal_send(mars->workload_queue,
			queue->push_wait_id[0]);

		/* flush id from push wait list */
		queue->push_wait_count--;
		memmove(&queue->push_wait_id[0], &queue->push_wait_id[1],
			sizeof(uint16_t) * queue->push_wait_count);
	}

	/* decrement queue count */
	queue->count--;

	/* increment queue pop ea */
	queue->pop_ea += queue->size;

	/* wrap to front of queue if necessary */
	if (queue->pop_ea == queue->buffer_ea + (queue->size * queue->depth))
		queue->pop_ea = queue->buffer_ea;
}

static int pop(struct mars_task_queue *queue, void *data, int peek, int try)
{
	/* check function params */
	if (!queue)
		return MARS_ERROR_NULL;
	if ((uintptr_t)queue & MARS_TASK_QUEUE_ALIGN_MASK)
		return MARS_ERROR_ALIGN;
	if (queue->direction != MARS_TASK_QUEUE_MPU_TO_HOST)
		return MARS_ERROR_STATE;

	mars_mutex_lock((struct mars_mutex *)queue);

	/* queue is empty so wait */
	while (!queue->count) {
		mars_mutex_unlock((struct mars_mutex *)queue);

		/* only try so return busy */
		if (try)
			return MARS_ERROR_BUSY;

		while (!queue->count)
			sched_yield();

		mars_mutex_lock((struct mars_mutex *)queue);
	}

	/* copy data from queue */
	memcpy(data, mars_ea_to_ptr(queue->pop_ea), queue->size);

	/* update queue data only if this is not a peek operation */
	if (!peek)
		pop_update(queue);

	mars_mutex_unlock((struct mars_mutex *)queue);

	return MARS_SUCCESS;
}

int mars_task_queue_pop(struct mars_task_queue *queue, void *data)
{
	return pop(queue, data, 0, 0);
}

int mars_task_queue_try_pop(struct mars_task_queue *queue, void *data)
{
	return pop(queue, data, 0, 1);
}

int mars_task_queue_peek(struct mars_task_queue *queue, void *data)
{
	return pop(queue, data, 1, 0);
}

int mars_task_queue_try_peek(struct mars_task_queue *queue, void *data)
{
	return pop(queue, data, 1, 1);
}
