/*
 * Copyright 2008 Sony Corporation of America
 *
 *
 * Permission is hereby granted, free of charge, to any person obtaining
 * a copy of this Library and associated documentation files (the
 * "Library"), to deal in the Library without restriction, including
 * without limitation the rights to use, copy, modify, merge, publish,
 * distribute, sublicense, and/or sell copies of the Library, and to
 * permit persons to whom the Library is furnished to do so, subject to
 * the following conditions:
 *
 *  The above copyright notice and this permission notice shall be
 *  included in all copies or substantial portions of the Library.
 *
 *  If you modify the Library, you may copy and distribute your modified
 *  version of the Library in object code or as an executable provided
 *  that you also do one of the following:
 *
 *   Accompany the modified version of the Library with the complete
 *   corresponding machine-readable source code for the modified version
 *   of the Library; or,
 *
 *   Accompany the modified version of the Library with a written offer
 *   for a complete machine-readable copy of the corresponding source
 *   code of the modified version of the Library.
 *
 *
 * THE LIBRARY IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND,
 * EXPRESS OR IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF
 * MERCHANTABILITY, FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT.
 * IN NO EVENT SHALL THE AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY
 * CLAIM, DAMAGES OR OTHER LIABILITY, WHETHER IN AN ACTION OF CONTRACT,
 * TORT OR OTHERWISE, ARISING FROM, OUT OF OR IN CONNECTION WITH THE
 * LIBRARY OR THE USE OR OTHER DEALINGS IN THE LIBRARY.
 */

#include <sched.h>
#include <unistd.h>
#include <string.h>
#include <malloc.h>

#include "mars/core.h"
#include "mars/mutex.h"
#include "mars/error.h"

#include "mutex_internal_types.h"
#include "kernel_internal_types.h"
#include "workload_queue.h"

int mars_workload_queue_create(struct mars_workload_queue **queue_ret)
{
	int block;
	int index;
	struct mars_workload_queue *queue;

	/* check function params */
	if (!queue_ret)
		return MARS_ERROR_NULL;

	/* allocate workload instance */
	queue = (struct mars_workload_queue *)memalign(
		MARS_WORKLOAD_QUEUE_ALIGN, sizeof(struct mars_workload_queue));
	if (!queue)
		return MARS_ERROR_MEMORY;

	/* initialize workload queue header */
	queue->header.flag = MARS_WORKLOAD_QUEUE_FLAG_NONE;
	queue->header.queue_ea = mars_ptr_to_ea(queue);
	queue->header.context_ea = mars_ptr_to_ea(&queue->context);

	/* initialize workload queue blocks */
	for (block = 0; block < MARS_WORKLOAD_NUM_BLOCKS; block++) {
		mars_mutex_reset((struct mars_mutex *)&queue->block[block]);

		for (index = 0; index < MARS_WORKLOAD_PER_BLOCK; index++) {
			uint64_t *bits = &queue->block[block].bits[index];

			MARS_BITS_SET(bits, TYPE,
				MARS_WORKLOAD_TYPE_NONE);
			MARS_BITS_SET(bits, STATE,
				MARS_WORKLOAD_STATE_NONE);
			MARS_BITS_SET(bits, PRIORITY,
				MARS_WORKLOAD_PRIORITY_MIN);
			MARS_BITS_SET(bits, COUNTER,
				MARS_WORKLOAD_COUNTER_MIN);
			MARS_BITS_SET(bits, SIGNAL,
				MARS_WORKLOAD_SIGNAL_OFF);
			MARS_BITS_SET(bits, WAIT_ID,
				MARS_WORKLOAD_ID_NONE);
		}
	}

	/* return workload queue instance */
	*queue_ret = queue;

	return MARS_SUCCESS;
}

int mars_workload_queue_destroy(struct mars_workload_queue *queue)
{
	int block;
	int index;
	uint16_t id = 0;

	/* check function params */
	if (!queue)
		return MARS_ERROR_NULL;

	/* check for any tasks left in workload queue */
	while (id < MARS_WORKLOAD_MAX) {
		block = id / MARS_WORKLOAD_PER_BLOCK;
		index = id % MARS_WORKLOAD_PER_BLOCK;

		if (MARS_BITS_GET(&queue->block[block].bits[index], STATE) !=
			MARS_WORKLOAD_STATE_NONE)
			return MARS_ERROR_STATE;

		id++;
	}

	/* free workload queue instance */
	free(queue);

	return MARS_SUCCESS;
}

int mars_workload_queue_exit(struct mars_workload_queue *queue)
{
	/* check function params */
	if (!queue)
		return MARS_ERROR_NULL;

	queue->header.flag = MARS_WORKLOAD_QUEUE_FLAG_EXIT;

	return MARS_SUCCESS;
}

int mars_workload_queue_add_begin(struct mars_workload_queue *queue,
				uint16_t *id, uint8_t type,
				struct mars_workload_context **workload)
{
	int block = 0;
	int index = 0;
	uint64_t *bits;

	/* check function params */
	if (!queue)
		return MARS_ERROR_NULL;
	if (!id)
		return MARS_ERROR_NULL;
	if (type != MARS_WORKLOAD_TYPE_TASK)
		return MARS_ERROR_PARAMS;

	mars_mutex_lock((struct mars_mutex *)&queue->block[block]);

	/* get bits from workload queue block */
	bits = &queue->block[block].bits[index];

	/* initialize return id to 0 */
	*id = 0;

	while (MARS_BITS_GET(bits, STATE) != MARS_WORKLOAD_STATE_NONE) {
		(*id)++;
		index++;
		if (index == MARS_WORKLOAD_PER_BLOCK) {
			index = 0;

			mars_mutex_unlock(
				(struct mars_mutex *)&queue->block[block]);

			if (++block == MARS_WORKLOAD_NUM_BLOCKS)
				return MARS_ERROR_LIMIT;

			mars_mutex_lock(
				(struct mars_mutex *)&queue->block[block]);
		}
		bits = &queue->block[block].bits[index];
	}

	/* check for valid state */
	if (MARS_BITS_GET(bits, STATE) != MARS_WORKLOAD_STATE_NONE) {
		mars_mutex_unlock((struct mars_mutex *)&queue->block[block]);
		return MARS_ERROR_STATE;
	}

	/* set type and set state to adding */
	MARS_BITS_SET(bits, TYPE, type);
	MARS_BITS_SET(bits, STATE, MARS_WORKLOAD_STATE_ADDING);

	mars_mutex_unlock((struct mars_mutex *)&queue->block[block]);

	/* if requested set workload context pointer to return */
	if (workload)
		*workload = &queue->context[*id];

	return MARS_SUCCESS;
}

int mars_workload_queue_add_end(struct mars_workload_queue *queue,
				uint16_t id)
{
	int block;
	int index;
	uint64_t *bits;

	/* check function params */
	if (!queue)
		return MARS_ERROR_NULL;
	if (id >= MARS_WORKLOAD_MAX)
		return MARS_ERROR_PARAMS;

	/* calculate block/index from id */
	block = id / MARS_WORKLOAD_PER_BLOCK;
	index = id % MARS_WORKLOAD_PER_BLOCK;

	mars_mutex_lock((struct mars_mutex *)&queue->block[block]);

	/* get bits from workload queue block */
	bits = &queue->block[block].bits[index];

	/* check for valid state */
	if (MARS_BITS_GET(bits, STATE) != MARS_WORKLOAD_STATE_ADDING) {
		mars_mutex_unlock((struct mars_mutex *)&queue->block[block]);
		return MARS_ERROR_STATE;
	}

	/* reset workload queue bits and set state to finished state */
	MARS_BITS_SET(bits, STATE, MARS_WORKLOAD_STATE_FINISHED);

	mars_mutex_unlock((struct mars_mutex *)&queue->block[block]);

	return MARS_SUCCESS;
}

int mars_workload_queue_add_cancel(struct mars_workload_queue *queue,
				uint16_t id)
{
	int block;
	int index;

	/* check function params */
	if (!queue)
		return MARS_ERROR_NULL;
	if (id >= MARS_WORKLOAD_MAX)
		return MARS_ERROR_PARAMS;

	/* calculate block/index from id */
	block = id / MARS_WORKLOAD_PER_BLOCK;
	index = id % MARS_WORKLOAD_PER_BLOCK;

	mars_mutex_lock((struct mars_mutex *)&queue->block[block]);

	/* check for valid state */
	if (MARS_BITS_GET(&queue->block[block].bits[index], STATE) !=
		MARS_WORKLOAD_STATE_ADDING) {
		mars_mutex_unlock((struct mars_mutex *)&queue->block[block]);
		return MARS_ERROR_STATE;
	}

	/* set state back to none state */
	MARS_BITS_SET(&queue->block[block].bits[index], STATE,
		MARS_WORKLOAD_STATE_NONE);

	mars_mutex_unlock((struct mars_mutex *)&queue->block[block]);

	return MARS_SUCCESS;
}

int mars_workload_queue_remove_begin(struct mars_workload_queue *queue,
				uint16_t id,
				struct mars_workload_context **workload)
{
	int block;
	int index;

	/* check function params */
	if (!queue)
		return MARS_ERROR_NULL;
	if (id >= MARS_WORKLOAD_MAX)
		return MARS_ERROR_PARAMS;

	/* calculate block/index from id */
	block = id / MARS_WORKLOAD_PER_BLOCK;
	index = id % MARS_WORKLOAD_PER_BLOCK;

	mars_mutex_lock((struct mars_mutex *)&queue->block[block]);

	/* check for valid state */
	if (MARS_BITS_GET(&queue->block[block].bits[index], STATE) !=
		MARS_WORKLOAD_STATE_FINISHED) {
		mars_mutex_unlock((struct mars_mutex *)&queue->block[block]);
		return MARS_ERROR_STATE;
	}

	/* set state to removing */
	MARS_BITS_SET(&queue->block[block].bits[index], STATE,
		MARS_WORKLOAD_STATE_REMOVING);

	mars_mutex_unlock((struct mars_mutex *)&queue->block[block]);

	/* if requested set workload context pointer to return */
	if (workload)
		*workload = &queue->context[id];

	return MARS_SUCCESS;
}

int mars_workload_queue_remove_end(struct mars_workload_queue *queue,
				uint16_t id)
{
	int block;
	int index;

	/* check function params */
	if (!queue)
		return MARS_ERROR_NULL;
	if (id >= MARS_WORKLOAD_MAX)
		return MARS_ERROR_PARAMS;

	/* calculate block/index from id */
	block = id / MARS_WORKLOAD_PER_BLOCK;
	index = id % MARS_WORKLOAD_PER_BLOCK;

	mars_mutex_lock((struct mars_mutex *)&queue->block[block]);

	/* check for valid state */
	if (MARS_BITS_GET(&queue->block[block].bits[index], STATE) !=
		MARS_WORKLOAD_STATE_REMOVING) {
		mars_mutex_unlock((struct mars_mutex *)&queue->block[block]);
		return MARS_ERROR_STATE;
	}

	/* set state to none */
	MARS_BITS_SET(&queue->block[block].bits[index], STATE,
		MARS_WORKLOAD_STATE_NONE);

	mars_mutex_unlock((struct mars_mutex *)&queue->block[block]);

	return MARS_SUCCESS;
}

int mars_workload_queue_remove_cancel(struct mars_workload_queue *queue,
				uint16_t id)
{
	int block;
	int index;

	/* check function params */
	if (!queue)
		return MARS_ERROR_NULL;
	if (id >= MARS_WORKLOAD_MAX)
		return MARS_ERROR_PARAMS;

	/* calculate block/index from id */
	block = id / MARS_WORKLOAD_PER_BLOCK;
	index = id % MARS_WORKLOAD_PER_BLOCK;

	mars_mutex_lock((struct mars_mutex *)&queue->block[block]);

	/* check for valid state */
	if (MARS_BITS_GET(&queue->block[block].bits[index], STATE) !=
		MARS_WORKLOAD_STATE_REMOVING) {
		mars_mutex_unlock((struct mars_mutex *)&queue->block[block]);
		return MARS_ERROR_STATE;
	}

	/* set state back to finished */
	MARS_BITS_SET(&queue->block[block].bits[index], STATE,
		MARS_WORKLOAD_STATE_FINISHED);

	mars_mutex_unlock((struct mars_mutex *)&queue->block[block]);

	return MARS_SUCCESS;
}

int mars_workload_queue_schedule_begin(struct mars_workload_queue *queue,
				uint16_t id, uint8_t priority,
				struct mars_workload_context **workload)
{
	int block;
	int index;

	/* check function params */
	if (!queue)
		return MARS_ERROR_NULL;
	if (id >= MARS_WORKLOAD_MAX)
		return MARS_ERROR_PARAMS;

	/* calculate block/index from id */
	block = id / MARS_WORKLOAD_PER_BLOCK;
	index = id % MARS_WORKLOAD_PER_BLOCK;

	mars_mutex_lock((struct mars_mutex *)&queue->block[block]);

	/* check for valid state */
	if (MARS_BITS_GET(&queue->block[block].bits[index], STATE) !=
		MARS_WORKLOAD_STATE_FINISHED) {
		mars_mutex_unlock((struct mars_mutex *)&queue->block[block]);
		return MARS_ERROR_STATE;
	}

	/* reset workload queue bits and set state to scheduling */
	MARS_BITS_SET(&queue->block[block].bits[index], STATE,
		MARS_WORKLOAD_STATE_SCHEDULING);
	MARS_BITS_SET(&queue->block[block].bits[index], PRIORITY,
		priority);
	MARS_BITS_SET(&queue->block[block].bits[index], COUNTER,
		MARS_WORKLOAD_COUNTER_MIN);
	MARS_BITS_SET(&queue->block[block].bits[index], SIGNAL,
		MARS_WORKLOAD_SIGNAL_OFF);
	MARS_BITS_SET(&queue->block[block].bits[index], WAIT_ID,
		MARS_WORKLOAD_ID_NONE);

	mars_mutex_unlock((struct mars_mutex *)&queue->block[block]);

	/* if requested set workload context pointer to return */
	if (workload)
		*workload = &queue->context[id];

	return MARS_SUCCESS;
}

int mars_workload_queue_schedule_end(struct mars_workload_queue *queue,
				uint16_t id)
{
	int block;
	int index;

	/* check function params */
	if (!queue)
		return MARS_ERROR_NULL;
	if (id >= MARS_WORKLOAD_MAX)
		return MARS_ERROR_PARAMS;

	/* calculate block/index from id */
	block = id / MARS_WORKLOAD_PER_BLOCK;
	index = id % MARS_WORKLOAD_PER_BLOCK;

	mars_mutex_lock((struct mars_mutex *)&queue->block[block]);

	/* check for valid state */
	if (MARS_BITS_GET(&queue->block[block].bits[index], STATE) !=
		MARS_WORKLOAD_STATE_SCHEDULING) {
		mars_mutex_unlock((struct mars_mutex *)&queue->block[block]);
		return MARS_ERROR_STATE;
	}

	/* set state to ready */
	MARS_BITS_SET(&queue->block[block].bits[index], STATE,
		MARS_WORKLOAD_STATE_READY);

	mars_mutex_unlock((struct mars_mutex *)&queue->block[block]);

	return MARS_SUCCESS;
}

int mars_workload_queue_schedule_cancel(struct mars_workload_queue *queue,
				uint16_t id)
{
	int block;
	int index;

	/* check function params */
	if (!queue)
		return MARS_ERROR_NULL;
	if (id >= MARS_WORKLOAD_MAX)
		return MARS_ERROR_PARAMS;

	/* calculate block/index from id */
	block = id / MARS_WORKLOAD_PER_BLOCK;
	index = id % MARS_WORKLOAD_PER_BLOCK;

	mars_mutex_lock((struct mars_mutex *)&queue->block[block]);

	/* check for valid state */
	if (MARS_BITS_GET(&queue->block[block].bits[index], STATE) !=
		MARS_WORKLOAD_STATE_SCHEDULING) {
		mars_mutex_unlock((struct mars_mutex *)&queue->block[block]);
		return MARS_ERROR_STATE;
	}

	/* set state back to finished */
	MARS_BITS_SET(&queue->block[block].bits[index], STATE,
		MARS_WORKLOAD_STATE_FINISHED);

	mars_mutex_unlock((struct mars_mutex *)&queue->block[block]);

	return MARS_SUCCESS;
}

int mars_workload_queue_wait(struct mars_workload_queue *queue,
				uint16_t id,
				struct mars_workload_context **workload)
{
	int block;
	int index;

	/* check function params */
	if (!queue)
		return MARS_ERROR_NULL;
	if (id >= MARS_WORKLOAD_MAX)
		return MARS_ERROR_PARAMS;

	/* calculate block/index from id */
	block = id / MARS_WORKLOAD_PER_BLOCK;
	index = id % MARS_WORKLOAD_PER_BLOCK;

	/* loop until workload state is finished */
	while (MARS_BITS_GET(&queue->block[block].bits[index], STATE) !=
		MARS_WORKLOAD_STATE_FINISHED) {
		if (MARS_BITS_GET(&queue->block[block].bits[index], STATE) ==
			MARS_WORKLOAD_STATE_NONE)
			return MARS_ERROR_STATE;
		sched_yield();
	}

	/* if requested set workload context pointer to return */
	if (workload)
		*workload = &queue->context[id];

	return MARS_SUCCESS;
}

int mars_workload_queue_try_wait(struct mars_workload_queue *queue,
				uint16_t id,
				struct mars_workload_context **workload)
{
	int block;
	int index;

	/* check function params */
	if (!queue)
		return MARS_ERROR_NULL;
	if (id >= MARS_WORKLOAD_MAX)
		return MARS_ERROR_PARAMS;

	/* calculate block/index from id */
	block = id / MARS_WORKLOAD_PER_BLOCK;
	index = id % MARS_WORKLOAD_PER_BLOCK;

	/* check for valid state */
	if (MARS_BITS_GET(&queue->block[block].bits[index], STATE) ==
		MARS_WORKLOAD_STATE_NONE)
		return MARS_ERROR_STATE;

	/* check if workload is finished */
	if (MARS_BITS_GET(&queue->block[block].bits[index], STATE) !=
		MARS_WORKLOAD_STATE_FINISHED)
		return MARS_ERROR_BUSY;

	/* if requested set workload context pointer to return */
	if (workload)
		*workload = &queue->context[id];

	return MARS_SUCCESS;
}

int mars_workload_queue_signal_send(struct mars_workload_queue *queue,
				uint16_t id)
{
	int block;
	int index;

	/* check function params */
	if (!queue)
		return MARS_ERROR_NULL;
	if (id >= MARS_WORKLOAD_MAX)
		return MARS_ERROR_PARAMS;

	/* calculate block/index from id */
	block = id / MARS_WORKLOAD_PER_BLOCK;
	index = id % MARS_WORKLOAD_PER_BLOCK;

	mars_mutex_lock((struct mars_mutex *)&queue->block[block]);

	/* check for valid state */
	if (MARS_BITS_GET(&queue->block[block].bits[index], STATE) ==
		MARS_WORKLOAD_STATE_NONE) {
		mars_mutex_unlock((struct mars_mutex *)&queue->block[block]);
		return MARS_ERROR_STATE;
	}

	/* set signal bit on */
	MARS_BITS_SET(&queue->block[block].bits[index], SIGNAL,
		MARS_WORKLOAD_SIGNAL_ON);

	mars_mutex_unlock((struct mars_mutex *)&queue->block[block]);

	return MARS_SUCCESS;
}
