/*
 * Copyright 2008 Sony Corporation of America
 *
 *
 * Permission is hereby granted, free of charge, to any person obtaining
 * a copy of this Library and associated documentation files (the
 * "Library"), to deal in the Library without restriction, including
 * without limitation the rights to use, copy, modify, merge, publish,
 * distribute, sublicense, and/or sell copies of the Library, and to
 * permit persons to whom the Library is furnished to do so, subject to
 * the following conditions:
 *
 *  The above copyright notice and this permission notice shall be
 *  included in all copies or substantial portions of the Library.
 *
 *  If you modify the Library, you may copy and distribute your modified
 *  version of the Library in object code or as an executable provided
 *  that you also do one of the following:
 *
 *   Accompany the modified version of the Library with the complete
 *   corresponding machine-readable source code for the modified version
 *   of the Library; or,
 *
 *   Accompany the modified version of the Library with a written offer
 *   for a complete machine-readable copy of the corresponding source
 *   code of the modified version of the Library.
 *
 *
 * THE LIBRARY IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND,
 * EXPRESS OR IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF
 * MERCHANTABILITY, FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT.
 * IN NO EVENT SHALL THE AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY
 * CLAIM, DAMAGES OR OTHER LIABILITY, WHETHER IN AN ACTION OF CONTRACT,
 * TORT OR OTHERWISE, ARISING FROM, OUT OF OR IN CONNECTION WITH THE
 * LIBRARY OR THE USE OR OTHER DEALINGS IN THE LIBRARY.
 */

#ifndef MARS_KERNEL_H
#define MARS_KERNEL_H

#include <stdlib.h>
#include <stdint.h>
#include <stddef.h>

#include "mars/mutex.h"
#include "mars/error.h"
#include "mars/task_types.h"

#include "kernel_internal_types.h"
#include "mutex_internal_types.h"
#include "workload_internal_types.h"
#include "task_internal_types.h"
#include "syscalls.h"
#include "dma.h"

#define MARS_KERNEL_STATUS_BUSY		0x0
#define MARS_KERNEL_STATUS_IDLE		0x1
#define MARS_KERNEL_STATUS_EXIT		0x2

typedef void (*mars_entry)(void *workload_args, void *kernel_syscalls);

#if defined(__cplusplus)
extern "C" {
#endif

void mars_reentry(void);

/* kernel scheduler function */
int scheduler(void);

/* kernel syscall functions */
uint64_t syscall_get_mars_context_ea(void);
uint32_t syscall_get_kernel_id(void);
uint16_t syscall_get_workload_id(void);
uint64_t syscall_get_workload_ea(void);
struct mars_workload_context *syscall_get_workload(void);
struct mars_workload_context *syscall_get_workload_by_id(uint16_t workload_id);

void syscall_init(void);
void syscall_exit(void);
void syscall_yield(void);
void syscall_resume(void);
int syscall_schedule(uint16_t workload_id, void *args, uint8_t priority);
int syscall_wait(uint16_t workload_id);
int syscall_try_wait(uint16_t workload_id);
int syscall_signal_send(uint16_t workload_id);
void syscall_signal_wait(void);
int syscall_signal_try_wait(void);

/* register saving / restoring functions */
void registers_save(void);
void registers_restore(void);

/* workload specific functions */
uint16_t workload_get_id(void);
uint64_t workload_get_ea(void);
struct mars_workload_context *workload_get(void);
struct mars_workload_context *workload_get_by_id(uint16_t workload_id);

void workload_exec(void);
void workload_exit(void);
void workload_yield(void);
void workload_resume(void);
int workload_schedule(uint16_t workload_id, void *args, uint8_t priority);
int workload_wait(uint16_t workload_id);
int workload_try_wait(uint16_t workload_id);
int workload_signal_send(uint16_t workload_id);
void workload_signal_wait(void);
int workload_signal_try_wait(void);

/* task specific functions */
void task_exec(struct mars_task_context *task);
void task_exit(struct mars_task_context *task);
void task_yield(struct mars_task_context *task);
void task_resume(struct mars_task_context *task);
void task_schedule(struct mars_task_context *task, struct mars_task_args *args);

extern struct mars_kernel_params kernel_params;
extern struct mars_kernel_syscalls kernel_syscalls;

#if defined(__cplusplus)
}
#endif

#endif
