/*
 * Copyright 2008 Sony Corporation of America
 *
 *
 * Permission is hereby granted, free of charge, to any person obtaining
 * a copy of this Library and associated documentation files (the
 * "Library"), to deal in the Library without restriction, including
 * without limitation the rights to use, copy, modify, merge, publish,
 * distribute, sublicense, and/or sell copies of the Library, and to
 * permit persons to whom the Library is furnished to do so, subject to
 * the following conditions:
 *
 *  The above copyright notice and this permission notice shall be
 *  included in all copies or substantial portions of the Library.
 *
 *  If you modify the Library, you may copy and distribute your modified
 *  version of the Library in object code or as an executable provided
 *  that you also do one of the following:
 *
 *   Accompany the modified version of the Library with the complete
 *   corresponding machine-readable source code for the modified version
 *   of the Library; or,
 *
 *   Accompany the modified version of the Library with a written offer
 *   for a complete machine-readable copy of the corresponding source
 *   code of the modified version of the Library.
 *
 *
 * THE LIBRARY IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND,
 * EXPRESS OR IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF
 * MERCHANTABILITY, FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT.
 * IN NO EVENT SHALL THE AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY
 * CLAIM, DAMAGES OR OTHER LIABILITY, WHETHER IN AN ACTION OF CONTRACT,
 * TORT OR OTHERWISE, ARISING FROM, OUT OF OR IN CONNECTION WITH THE
 * LIBRARY OR THE USE OR OTHER DEALINGS IN THE LIBRARY.
 */

#include "kernel.h"

extern struct mars_workload_queue_header queue_header;
extern struct mars_workload_queue_block queue_block;
extern struct mars_workload_context workload;
extern uint8_t workload_type;
extern uint8_t workload_state;
extern uint16_t workload_index;
extern uint64_t workload_ea;

uint16_t workload_get_id(void)
{
	return workload_index;
}

uint64_t workload_get_ea(void)
{
	return workload_ea;
}

struct mars_workload_context *workload_get(void)
{
	return &workload;
}

struct mars_workload_context *workload_get_by_id(uint16_t workload_id)
{
	static struct mars_workload_context requested_workload;
	uint64_t requested_workload_ea;

	/* check function params */
	if (workload_id >= MARS_WORKLOAD_MAX)
		return NULL;

	/* workload id is same as current workload so return local copy */
	if (workload_id == workload_index)
		return &workload;

	/* get ea of workload to get */
	requested_workload_ea = queue_header.context_ea +
		workload_id * sizeof(struct mars_workload_context);

	/* dma the workload context code into LS from main memory */
	mars_dma_get_and_wait((void *)&requested_workload,
		requested_workload_ea, sizeof(struct mars_workload_context),
		MARS_DMA_TAG);

	return &requested_workload;
}

void workload_exec(void)
{
	/* set current workload state to running */
	workload_state = MARS_WORKLOAD_STATE_RUNNING;

	/* workload type specific handling */
	switch (workload_type) {
	case MARS_WORKLOAD_TYPE_TASK:
		task_exec((struct mars_task_context *)&workload);
		break;
	}
}

void workload_exit(void)
{
	/* set current workload state to finished */
	workload_state = MARS_WORKLOAD_STATE_FINISHED;

	/* workload type specific handling */
	switch (workload_type) {
	case MARS_WORKLOAD_TYPE_TASK:
		task_exit((struct mars_task_context *)&workload);
		break;
	}
}

void workload_yield(void)
{
	/* set current workload state to ready */
	workload_state = MARS_WORKLOAD_STATE_READY;

	/* workload type specific handling */
	switch (workload_type) {
	case MARS_WORKLOAD_TYPE_TASK:
		/* save non-volatile registers */
		registers_save();

		/* yield the task context */
		task_yield((struct mars_task_context *)&workload);
		break;
	}
}

void workload_resume(void)
{
	/* set current workload state to running */
	workload_state = MARS_WORKLOAD_STATE_RUNNING;

	/* workload type specific handling */
	switch (workload_type) {
	case MARS_WORKLOAD_TYPE_TASK:
		/* resume the task context */
		task_resume((struct mars_task_context *)&workload);

		/* restore non-volatile registers */
		registers_restore();
		break;
	}
}

int workload_schedule(uint16_t workload_id, void *args, uint8_t priority)
{
	struct mars_workload_context schedule_workload;
	uint8_t schedule_workload_type;
	uint64_t schedule_workload_ea;
	uint64_t block_ea;
	int block;
	int index;

	/* check function params */
	if (workload_id >= MARS_WORKLOAD_MAX)
		return MARS_ERROR_PARAMS;
	if (workload_id == workload_index)
		return MARS_ERROR_PARAMS;

	/* calculate block/index from id */
	block = workload_id / MARS_WORKLOAD_PER_BLOCK;
	index = workload_id % MARS_WORKLOAD_PER_BLOCK;

	/* calculate block ea */
	block_ea = queue_header.queue_ea +
		offsetof(struct mars_workload_queue, block) +
		sizeof(struct mars_workload_queue_block) * block;

	mars_mutex_lock_get(block_ea, (struct mars_mutex *)&queue_block);

	/* make sure workload is in the correct state */
	if (MARS_BITS_GET(&queue_block.bits[index], STATE) !=
		MARS_WORKLOAD_STATE_FINISHED) {
		mars_mutex_unlock_put(block_ea,
			(struct mars_mutex *)&queue_block);
		return MARS_ERROR_STATE;
	}

	/* get information of workload to schedule */
	schedule_workload_type = MARS_BITS_GET(&queue_block.bits[index], TYPE);
	schedule_workload_ea = queue_header.context_ea +
		workload_id * sizeof(struct mars_workload_context);

	/* dma the workload context code into LS from main memory */
	mars_dma_get_and_wait((void *)&schedule_workload,
		schedule_workload_ea, sizeof(struct mars_workload_context),
		MARS_DMA_TAG);

	/* workload type specific handling */
	switch (schedule_workload_type) {
	case MARS_WORKLOAD_TYPE_TASK:
		task_schedule((struct mars_task_context *)&schedule_workload,
			(struct mars_task_args *)args);
		break;
	}

	/* dma the workload context code into main memory from LS */
	mars_dma_put_and_wait((void *)&schedule_workload,
		schedule_workload_ea, sizeof(struct mars_workload_context),
		MARS_DMA_TAG);

	/* initialize queue block bits */
	MARS_BITS_SET(&queue_block.bits[index], STATE,
		MARS_WORKLOAD_STATE_READY);
	MARS_BITS_SET(&queue_block.bits[index], PRIORITY,
		priority);
	MARS_BITS_SET(&queue_block.bits[index], COUNTER,
		MARS_WORKLOAD_COUNTER_MIN);
	MARS_BITS_SET(&queue_block.bits[index], SIGNAL,
		MARS_WORKLOAD_SIGNAL_OFF);
	MARS_BITS_SET(&queue_block.bits[index], WAIT_ID,
		MARS_WORKLOAD_ID_NONE);

	mars_mutex_unlock_put(block_ea, (struct mars_mutex *)&queue_block);

	return MARS_SUCCESS;
}

int workload_wait(uint16_t workload_id)
{
	uint64_t block_ea;
	int block;
	int index;

	/* check function params */
	if (workload_id >= MARS_WORKLOAD_MAX)
		return MARS_ERROR_PARAMS;
	if (workload_id == workload_index)
		return MARS_ERROR_PARAMS;

	/* calculate block/index from id */
	block = workload_index / MARS_WORKLOAD_PER_BLOCK;
	index = workload_index % MARS_WORKLOAD_PER_BLOCK;

	/* calculate block ea */
	block_ea = queue_header.queue_ea +
		offsetof(struct mars_workload_queue, block) +
		sizeof(struct mars_workload_queue_block) * block;

	mars_mutex_lock_get(block_ea, (struct mars_mutex *)&queue_block);

	/* make sure workload is initialized */
	if (MARS_BITS_GET(&queue_block.bits[index], STATE) ==
		MARS_WORKLOAD_STATE_NONE) {
		mars_mutex_unlock_put(block_ea,
			(struct mars_mutex *)&queue_block);
		return MARS_ERROR_STATE;
	}

	/* set the workload id to wait for */
	MARS_BITS_SET(&queue_block.bits[index], WAIT_ID, workload_id);

	mars_mutex_unlock_put(block_ea, (struct mars_mutex *)&queue_block);

	return MARS_SUCCESS;
}

int workload_try_wait(uint16_t workload_id)
{
	uint64_t block_ea;
	int block;
	int index;

	/* check function params */
	if (workload_id >= MARS_WORKLOAD_MAX)
		return MARS_ERROR_PARAMS;
	if (workload_id == workload_index)
		return MARS_ERROR_PARAMS;

	/* calculate block/index from id */
	block = workload_id / MARS_WORKLOAD_PER_BLOCK;
	index = workload_id % MARS_WORKLOAD_PER_BLOCK;

	/* calculate block ea */
	block_ea = queue_header.queue_ea +
		offsetof(struct mars_workload_queue, block) +
		sizeof(struct mars_workload_queue_block) * block;

	mars_mutex_lock_get(block_ea, (struct mars_mutex *)&queue_block);

	/* check for valid state */
	if (MARS_BITS_GET(&queue_block.bits[index], STATE) ==
		MARS_WORKLOAD_STATE_NONE) {
		mars_mutex_unlock_put(block_ea,
			(struct mars_mutex *)&queue_block);
		return MARS_ERROR_STATE;
	}

	mars_mutex_unlock_put(block_ea, (struct mars_mutex *)&queue_block);

	/* check if workload is finished */
	if (MARS_BITS_GET(&queue_block.bits[index], STATE) !=
		MARS_WORKLOAD_STATE_FINISHED)
		return MARS_ERROR_BUSY;

	return MARS_SUCCESS;
}

int workload_signal_send(uint16_t workload_id)
{
	uint64_t block_ea;
	int block;
	int index;

	/* check function params */
	if (workload_id >= MARS_WORKLOAD_MAX)
		return MARS_ERROR_PARAMS;
	if (workload_id == workload_index)
		return MARS_ERROR_PARAMS;

	/* calculate block/index from id */
	block = workload_id / MARS_WORKLOAD_PER_BLOCK;
	index = workload_id % MARS_WORKLOAD_PER_BLOCK;

	/* calculate block ea */
	block_ea = queue_header.queue_ea +
		offsetof(struct mars_workload_queue, block) +
		sizeof(struct mars_workload_queue_block) * block;

	mars_mutex_lock_get(block_ea, (struct mars_mutex *)&queue_block);

	/* make sure workload is initialized */
	if (MARS_BITS_GET(&queue_block.bits[index], STATE) ==
		MARS_WORKLOAD_STATE_NONE) {
		mars_mutex_unlock_put(block_ea,
			(struct mars_mutex *)&queue_block);
		return MARS_ERROR_STATE;
	}

	/* set the workload signal */
	MARS_BITS_SET(&queue_block.bits[index], SIGNAL,
		MARS_WORKLOAD_SIGNAL_ON);

	mars_mutex_unlock_put(block_ea, (struct mars_mutex *)&queue_block);

	return MARS_SUCCESS;
}

void workload_signal_wait(void)
{
	/* set current workload state to waiting */
	workload_state = MARS_WORKLOAD_STATE_WAITING;

	/* workload type specific handling */
	switch (workload_type) {
	case MARS_WORKLOAD_TYPE_TASK:
		/* save non-volatile registers */
		registers_save();

		/* yield the task context */
		task_yield((struct mars_task_context *)&workload);
		break;
	}
}

int workload_signal_try_wait(void)
{
	uint64_t block_ea;
	int block;
	int index;

	/* check function params */
	if (workload_index >= MARS_WORKLOAD_MAX)
		return MARS_ERROR_PARAMS;

	/* calculate block/index from id */
	block = workload_index / MARS_WORKLOAD_PER_BLOCK;
	index = workload_index % MARS_WORKLOAD_PER_BLOCK;

	/* calculate block ea */
	block_ea = queue_header.queue_ea +
		offsetof(struct mars_workload_queue, block) +
		sizeof(struct mars_workload_queue_block) * block;

	mars_mutex_lock_get(block_ea, (struct mars_mutex *)&queue_block);

	/* check for valid state */
	if (MARS_BITS_GET(&queue_block.bits[index], STATE) ==
		MARS_WORKLOAD_STATE_NONE) {
		mars_mutex_unlock_put(block_ea,
			(struct mars_mutex *)&queue_block);
		return MARS_ERROR_STATE;
	}

	mars_mutex_unlock_put(block_ea, (struct mars_mutex *)&queue_block);

	/* return busy if task has not received signal */
	if (MARS_BITS_GET(&queue_block.bits[index], SIGNAL) !=
		MARS_WORKLOAD_SIGNAL_ON)
		return MARS_ERROR_BUSY;

	return MARS_SUCCESS;
}
